//
//  ENAccountContext.h
//  CoreNote
//
//  Created by Steve White on 5/15/12.
//  Copyright (c) 2012 Evernote Corporation. All rights reserved.
//

#ifndef ENAccountContext_HEADER_GUARD
#define ENAccountContext_HEADER_GUARD

#import <Foundation/Foundation.h>

#import "ENCoreDataLocalStore.h"
#import "ENCoreDataLocalStoreBackingDelegate.h"
#import "ENSyncEngine.h"

@class EDAMClientCache;
@class EDAMSharedNotebookHTTPClient;
@class ENCacheManager;
@class ENPreviewManager;
@class ENReverseGeocodingManager;
@class ENSegmentedDataCache;
@class ENSyncEngine;
@class ENSyncState;
@class EDAMHTTPClient;
@class EDAMUsage;
@protocol ENSearchStore;
@protocol ENMetadataStore;
@protocol ENSyncEngineConflictDelegate;
@class NSManagedObjectID;


typedef enum {
  ENAccountContextSaveType_None,
  ENAccountContextSaveType_Now,
  ENAccountContextSaveType_Idle
} ENAccountContextSaveType;


typedef void (^ENAccountContextLocalStoreReadOnlyWorkerBlock)(ENCoreDataLocalStore *localStore);
typedef void (^ENAccountContextLocalStoreMOIDReadOnlyWorkerBlock)(ENCoreDataLocalStore *localStore, id managedObject);
typedef ENAccountContextSaveType (^ENAccountContextLocalStoreWorkerBlock)(ENCoreDataLocalStore *localStore);
typedef ENAccountContextSaveType (^ENAccountContextLocalStoreMOIDWorkerBlock)(ENCoreDataLocalStore *localStore, id managedObject);
typedef void (^ENAccountContextLocalStoreCompletionBlock)(BOOL success, NSError *error);
typedef void (^ENAccountContextEmailVerificationCompletionBlock)(BOOL success);
typedef void (^ENAccountContextNoteCompletionBlock)(ENNoteMOID * noteID, NSError * error);
typedef void (^ENAccountContextNotebookCompletionBlock)(ENNotebookMOID * notebookID, NSError * error);
typedef void (^ENAccountContextPostSyncBlock)(ENAccountContext * context, ENSyncEngine * syncEngine, NSArray * errors);


// Notification posted by ENAccountContext if service reachability changes
// (Only sent if -sendsServiceReachabilityNotifications return YES)
extern NSString *const ENAccountContextServiceReachabilityDidChangeNotification;

@interface ENAccountContext : NSObject <ENCoreDataLocalStoreBackingDelegate> {
  BOOL _anonymousContext;
  
  ENCoreDataLocalStore *_readOnlyLocalStore;
  ENCoreDataLocalStore * _readWriteLocalStore;
  
  NSArray *_directoryPathComponents;
  NSString *_cacheDirectory;
  NSString *_documentsDirectory;
  NSString *_contentDirectory;
  BOOL _unifiedFileSystem;
  
  EDAMUsage *_sessionTracker;
  ENSyncEngine *_syncEngine;
  ENSyncState *_syncState;
  int32_t _automaticSyncDisabled;
  
  EDAMClientCache *_edamClientCache;
  EDAMHTTPClient *_edamClient;
  ENCacheManager *_cacheManager;
  
  ENReverseGeocodingManager *_geocodingManager;
  
  ENPreviewManager *_previewManager;
  ENSegmentedDataCache *_dataCache;
}

@property (readonly, nonatomic) EDAMUser *userInfo;
@property (readonly,strong) EDAMBootstrapProfile * profile;

@property (strong, nonatomic, readonly) NSNumber *businessID;

@property (readonly, nonatomic) ENCoreDataLocalStore *readOnlyLocalStore;
@property (readonly, nonatomic) ENCoreDataLocalStore *readWriteLocalStore;
@property (readonly, nonatomic) id<ENSearchStore> searchStore;
@property (readonly, nonatomic) id<ENMetadataStore> metadataStore;

@property (readonly, nonatomic) ENSyncEngine *syncEngine;
@property (weak, nonatomic) id<ENSyncEngineConflictDelegate> syncEngineConflictDelegate;
@property (readonly, nonatomic) ENSyncState *syncState;
@property (readonly, nonatomic) EDAMHTTPClient *edamClient;
@property (readonly, nonatomic) EDAMHTTPClient *edamNoteStoreClientForBusiness;
@property (readonly, nonatomic) EDAMClientCache *edamClientCache;
@property (assign) NSTimeInterval autoSyncInterval;


- (EDAMHTTPClient *) edamNoteStoreClientForNotebookGUID:(NSString *)notebookGUID;
- (EDAMHTTPClient *) edamNoteStoreClientForObjectID: (NSManagedObjectID *) objectID;

@property (readonly, nonatomic) ENCacheManager *cacheManager;
- (BOOL) shouldUseSessionDownloader;

@property (readonly, nonatomic) NSString* serviceHost;

// Indicates that the service is reachable
// Do not observe this property. Register for the ENAccountContextServiceReachabilityDidChangeNotification instead.
@property (readonly) BOOL serviceReachable;

// If set to true, the account posts a notification when reachability changes. Default is NO.
@property (nonatomic) BOOL sendsServiceReachabilityNotifications;

@property (readonly, nonatomic) NSString *activityLogFile;

@property (readonly, nonatomic) BOOL hasConfiguredGeocodingManager;
@property (readonly, nonatomic) ENReverseGeocodingManager *geocodingManager;

@property (readonly, nonatomic) ENPreviewManager *previewManager;
@property (readonly, nonatomic) ENSegmentedDataCache *dataCache;

@property (nonatomic, readonly, getter=isPremiumAccount) BOOL premiumAccount;

// One could think of this as an "is activated business account"
// It returns YES only if the user is part of a business, and if
// SSO is required that they've completed the SSO requirements.
@property (nonatomic, readonly, getter=isBusinessAccount) BOOL businessAccount;

// isBusinessAccount will return NO if the user is part of a business,
// but hasn't yet done the necessary SSO.  This method will return
// YES in that scenario, and NO in all others.
- (BOOL) needsBusinessSSO;

// Returns YES if SSO authentication has completed once.
- (BOOL) businessSSOHasCompletedOnce;

+ (NSString *) baseContentPath;

+ (void) setBaseDocumentPath:(NSString *)documentsPath;
+ (void) setBaseCachePath:(NSString *)cachePath;

- (void) disableAutomaticSync;
- (void) enableAutomaticSync;
- (BOOL) isAutomaticSyncEnabled;

#if defined ANONYMOUS_CONTEXT_SUPPORT && ANONYMOUS_CONTEXT_SUPPORT != 0
- (id) initForAnonymousUser;
- (BOOL) isAnonymousContext;
- (void) convertAnonymousWithLongLivedAuthResult:(EDAMAuthenticationResult *)authResult
                                   onServiceHost:(NSString *)serviceHost;
#endif

#if defined APPLE_INSTORE_DEMO && APPLE_INSTORE_DEMO != 0
- (id) initForAppleStoreDemo;
#endif

- (id) initWithLongLivedAuthResult:(EDAMAuthenticationResult *)authResult
                       serviceHost:(NSString *)serviceHost
                 unifiedFileSystem:(BOOL)unifiedFileSystem;


- (id) initWithLongLivedAuthResult:(EDAMAuthenticationResult *)authResult
                           profile:(EDAMBootstrapProfile*)profile
                 unifiedFileSystem:(BOOL)unifiedFileSystem;


- (id) initWithUsername:(NSString *)username
               password:(NSString *)password
						serviceHost:(NSString *)serviceHost;

- (id) initWithUsername:(NSString *)username
               password:(NSString *)password
						serviceHost:(NSString *)serviceHost
      unifiedFileSystem:(BOOL)unifiedFileSystem;

- (void) closeContext;
- (void) logout;
- (void) sync;
- (void) backgroundSessionDidResume;

- (NSString *) cacheDirectory;
- (NSString *) documentsDirectory;
- (NSString *) contentDirectory;

- (void)verifyEmailAddress:(ENAccountContextEmailVerificationCompletionBlock)completion;
- (void)confirmEmailAddress:(ENAccountContextEmailVerificationCompletionBlock)completion;

//
// Local store helper methods
//

// Asynchronous read only blocks, available to call from any thread.
// Attempts to invoke localStore save: will result in an assertion.
- (void)performLocalStoreReadBlock:(ENAccountContextLocalStoreReadOnlyWorkerBlock)block;
- (void)performWithObjectID:(NSManagedObjectID *)moid
        localStoreReadBlock:(ENAccountContextLocalStoreMOIDReadOnlyWorkerBlock)block;

// Synchronous read only blocks, available to call from any thread.
// Attempts to invoke localStore save: will result in an assertion.
- (void)performLocalStoreReadBlockAndWait:(ENAccountContextLocalStoreReadOnlyWorkerBlock)block;
- (void)performWithObjectID:(NSManagedObjectID *)moid
 localStoreReadBlockAndWait:(ENAccountContextLocalStoreMOIDReadOnlyWorkerBlock)block;

// Asynchronous read write blocks, available to call from any thread.
- (void)performLocalStoreReadWriteBlock:(ENAccountContextLocalStoreWorkerBlock)block
                             completion:(ENAccountContextLocalStoreCompletionBlock)completionBlock;
- (void)performWithObjectID:(NSManagedObjectID *)moid
   localStoreReadWriteBlock:(ENAccountContextLocalStoreMOIDWorkerBlock)block
                 completion:(ENAccountContextLocalStoreCompletionBlock)completionBlock;

// Synchronous read write blocks, *cannot* be called from the main thread.
- (void)performLocalStoreReadWriteBlockAndWait:(ENAccountContextLocalStoreWorkerBlock)block
                                    completion:(ENAccountContextLocalStoreCompletionBlock)completionBlock;
- (void)performWithObjectID:(NSManagedObjectID *)moid
localStoreReadWriteBlockAndWait:(ENAccountContextLocalStoreMOIDWorkerBlock)block
                 completion:(ENAccountContextLocalStoreCompletionBlock)completionBlock;


/**
 * Unique ID for this context
 */
@property (readonly) NSString * contextID;


/**
 * Construct our read-write localstore.  Handy for overriding by subclasses.
 * In fact, really only to be used by subclasses.
 */
- (ENCoreDataLocalStore *) instantiateReadWriteLocalStore;


/**
 * Synchronous method to update the sharing for a given note.  All specified EDAM contacts will be given
 * the specified privilege unless they already have a higher one.  Must be called on bg thread.
 */
- (BOOL) updateSharingForNoteWithGUID: (NSString *) noteGUID
                       privilegeLevel: (enum EDAMSharedNotePrivilegeLevel) privilegeLevel
                           recipients: (NSArray *) contactArray
                                error: (NSError * __autoreleasing *) outError;


/**
 *  Fetch the latest version of the specified note from the service.
 *  This results in a fetch of the note from the service, so don't call this unless you 
 *  know the note is out-of-date. e.g. from a noteLockStatus call.
 *
 *  @param noteGUID         the note guid
 *  @param callerCompletion completion block, called on main thread, always called
 */
- (void) refreshNoteFromService: (ENNoteMO *) note
                 mainCompletion: (ENAccountContextNoteCompletionBlock) callerCompletion;


/**
 *  "Resolve" the note with the specified guid.  That means, find it in the local DB or,
 *  if not present try to fetch is from the server and add it to the local store.
 *
 *  @param noteGUID     the note GUID
 *  @param shardID      the note's shard ID
 *  @param noteStoreURL absolute url for note store in string form
 *  @param completion   completion block, always called, called on main thread
 */
- (void) resolveNoteWithGUID: (NSString *) noteGUID
                       title: (NSString *) noteTitle
                     shardID: (NSString *) shardID
                noteStoreURL: (NSString *) noteStoreURLString
              mainCompletion: (ENAccountContextNoteCompletionBlock) completion;


/**
 *  "Resolve" the notebook with the specified guid.  That means, if not in the local db,
 *  we attempt to sync it.
 *
 *  @param notebookGUID       notebook GUID
 *  @param noteStoreURLString absolute notestore url of notebook in string form
 *  @param completion         completion block, always called, called on main thread
 */
- (void) resolveNotebookWithGUID: (NSString *) notebookGUID
                    noteStoreURL: (NSString *) noteStoreURLString
                  mainCompletion: (ENAccountContextNotebookCompletionBlock) completion;


/**
 * Queue up a block to run when sync completes, and kick off a sync.
 * If a sync is currently in progress, a second sync wil be kicked off and
 * then the block will be executed.
 *
 * NB: BLOCK WILL BE EXECUTED ON GLOBAL BG QUEUE.
 */
- (void) syncAndPerformBlock: (ENAccountContextPostSyncBlock) block;


#if !TARGET_OS_IPHONE

/**
 * Determine how many accounts are stored locally for the given profile
 */
+ (NSUInteger) countOfLocalAccountsForProfile: (EDAMBootstrapProfile *) profile;


#endif //!TARGET_OS_IPHONE


/**
 *
 * ===================================================================================================
 * YOU SHOULDN'T NEED TO USE THE FOLLOWING METHODS UNLESS YOU ARE SUBCLASSING OR DOING SOMETHING ELITE
 * ===================================================================================================
 *
 */
- (ENSyncState *) reconstituteSyncState;


@end


#endif