//
//  ENCoreDataLocalStore.h
//  CoreNote
//
//  Created by Steve White on 10/12/11.
//  Copyright 2011 Evernote Corporation. All rights reserved.
//

#if !TARGET_OS_IPHONE
# import <AppKit/AppKit.h>
#endif

#import <CoreData/CoreData.h>
#import <Foundation/Foundation.h>

#import "ENBusinessNotebookMO.h"
#import "ENClientPreferenceMO.h"
#import "ENExpungeMO.h"
#import "ENLinkedNotebookMO.h"
#import "ENNoteMO.h"
#import "ENNotebookMO.h"

#if !TARGET_OS_IPHONE
#import "ENNotebookMO+Permissions.h"
#endif

#import "ENPublishingMO.h"
#import "ENResourceMO.h"
#import "ENSavedSearchMO.h"
#import "ENSharedNotebookMO.h"
#import "ENTagMO.h"
#import "ENUserProfile.h"
#import "ENUserProfileMO.h"

#import "ENLegacyObjectMapper.h"
#import "ENMetadataStore.h"
#import "ENSearchStore.h"

extern NSString * const ENLocalStoreValidationErrorDomain;
extern NSString * const ENLocalStoreValidationSuggestionKey;
extern NSString * const ENLocalStoreDuplicateObjectKey;
extern NSString * const ENLocalStoreDidSaveNotification;
extern NSString * const ENLocalStoreHasDirtyEntities;
extern NSString * const ENLocalStoreErrorDomain;

extern NSString * const ENLinkedNotebookInvalidGUID;
extern NSString * const ENDanglingSharedNotesNotebookGUID;
extern NSString * const ENDanglingSharedNotesAccountID;


enum {
  ENLocalStoreNameTooLongError = 1000,
  ENLocalStoreNameTooShortError,
  ENLocalStoreNameMustBeUniqueError,
  ENLocalStoreNoPersistentStoreError,
  ENLocalStoreCouldntFindObjectError,
};

@class ENAccountContext;
@protocol ENCoreDataLocalStoreBackingDelegate;


@interface ENCoreDataLocalStore : NSObject 

@property (strong, nonatomic) NSNumber *businessID;

@property (strong, nonatomic) NSString *storeDataDirectory;
@property (strong, nonatomic) NSString *noteContentDirectory;

@property (strong, nonatomic) NSString *pendingNotesDirectory;

@property (strong, nonatomic) NSString *storePath;
@property (strong, nonatomic) NSString *managedObjectModelPath;

@property (strong, nonatomic) NSManagedObjectContext *managedObjectContext;
@property (strong, nonatomic) id<ENSearchStore> searchStore;
@property (strong, nonatomic) id<ENMetadataStore> metadataStore;
@property (strong, nonatomic) id<ENLegacyObjectMapper> objectMapper;

@property (nonatomic, readonly) CFAbsoluteTime lastSaved;
@property (nonatomic, readonly) uint32_t crudCount;


/** Indicates if this local store allows modifications to its managed objects */
@property (nonatomic,readonly) BOOL writable;
@property (readonly,nonatomic) BOOL hasLocalNotebooks;

// reference to our master account context
@property (weak) ENAccountContext * accountContext;

@property (weak, nonatomic) id<ENCoreDataLocalStoreBackingDelegate> backingDelegate;

#pragma mark -
#pragma mark
- (NSPersistentStore *) addStoreWithFileAtPath:(NSString *)path
                                         error:(NSError **)outError;
- (BOOL) removeStoreForFileAtPath:(NSString *)path
                            error:(NSError **)outError;


- (ENCoreDataLocalStore *) threadSafeLocalStore;
- (ENCoreDataLocalStore *) newThreadSafeLocalStore;
- (ENCoreDataLocalStore *) filteredLocalStoreWithAccountID:(NSString *)accountID;

#pragma mark -
#pragma mark ObjectID helpers
- (BOOL) isObjectRegisteredInStore:(id)object;
- (BOOL) isObjectIDRegisteredInStore:(id)object;
- (id) objectWithID:(id)objectID;
- (void) expungeObject:(id)object; // warning: this should be used very sparingly and only in certain rare circumstances (refer to Geech or Steve if you're not sure)
/**
 *  Return the specified objects by their object ID, this assumes the object ID exists and never returns nil.
 *
 *  This is meant for convenience and for large arrays you will get better performance with
 *     [NSPredicate predicateWithFormat:@"SELF in %@", objectIDs]
 *
 *  @param objectIDs the core data object IDs
 *
 *  @return the core data objects corresponding to the specified IDs
 */
- (NSArray *) objectsWithIDs:(NSArray *)objectIDs;
- (id) objectWithIDString:(NSString *)string;
- (NSManagedObjectID *) objectIDWithEvernoteAppURL: (NSURL*) appURL;
- (id) objectWithEvernoteAppURL:(NSURL *)appURL;

#pragma mark -
#pragma mark Save helpers
- (void) reset;
- (void) resetAndPropagate:(BOOL)propagate;
- (BOOL) hasUnsavedChanges;
- (BOOL) save:(NSError *__autoreleasing *)outError;
- (void) saveWhenIdle;
- (void) rollback;

#pragma mark -
#pragma mark 
- (void)performBlock:(void (^)())block;
- (void)performBlockAndWait:(void (^)())block;

- (void) usingObject:(NSManagedObject *)object
        performBlock:(void(^)(ENCoreDataLocalStore *localStore, NSManagedObject *object))block
          completion:(void(^)(BOOL success, NSError *error))completionBlock;

- (void) createObjectOfType:(NSString *)entityType
                 usingBlock:(void(^)(ENCoreDataLocalStore *localStore, NSManagedObject *object))block
                 completion:(void(^)(BOOL success, NSError *error))completionBlock;

#pragma mark -
#pragma mark 
- (BOOL) hasDirtyEntities;
- (NSPredicate *) dirtyPredicate;

- (void) deleteServiceObject:(id)serviceObject;
- (ENExpungeMO *) findOrCreateExpungeRecordForEntity:(NSString *)entityName
                                            withGuid:(NSString *)guid;

/**
 * Predicate for fetching default notebook
 */
- (NSPredicate *) defaultNotebookPredicate;


#pragma mark -
#pragma mark Client preferences
- (ENClientPreferenceMO *) clientPreferenceForKey:(NSString *)key;
- (NSArray *) clientPreferencesWithKeys:(NSArray *)keys;
- (ENClientPreferenceMO *) createClientPreference;
- (ENClientPreferenceMO *) findOrCreateClientPreferenceWithKey:(NSString *)key;
- (NSArray *) dirtyClientPreferenceIDs;
- (NSDictionary *) mapOfClientPreferencesWithKeys:(NSArray *)keys;

#pragma mark -
#pragma mark Linked notebooks
- (NSArray *) dirtyLinkedNotebookIDs;
- (NSArray *) linkedNotebookGUIDsToExpunge;

- (NSUInteger) numberOfLinkedNotebooks;

- (ENLinkedNotebookMO *) linkedNotebookWithOwnerName:(NSString *)ownerName
                                                 URI:(NSString *)URI;
- (ENLinkedNotebookMO *) linkedNotebookWithShareKey:(NSString *)shareKey;
- (ENLinkedNotebookMO *) linkedNotebookWithGUID:(NSString *)guid;
- (ENLinkedNotebookMO *) linkedNotebookWithName:(NSString *)name;
- (NSArray *) linkedNotebooksInStackNamed:(NSString *)stackName;
- (NSArray *) linkedNotebooksWithGUIDs:(NSArray *)guids;
- (NSArray *) linkedNotebooksWithOwnerID:(uint32_t)ownerID;
- (NSArray *) linkedNotebooksWithBusinessID:(NSNumber *)businessID;
- (NSArray *) allLinkedNotebooks;

- (ENLinkedNotebookMO *) createLinkedNotebook;
- (void) deleteLinkedNotebook:(ENLinkedNotebookMO *)linkedNotebook;
- (void) deleteLinkedNotebooks:(NSArray *)linkedNotebooks;

- (NSDictionary *) mapOfLinkedNotebooksWithGUIDs:(NSArray *)guids;

- (id) newLinkedNotebooksControllerWithPredicate:(NSPredicate *)predicate
                                 sortDescriptors:(NSArray *)sortDescriptors;

- (BOOL) validateName:(id*)value
    forLinkedNotebook:(ENLinkedNotebookMO *)notebook
                error:(NSError *__autoreleasing *)error;

#pragma mark -
#pragma mark Notes
- (BOOL) hasNotes;
- (NSArray *) dirtyNoteIDs;
- (NSArray *) allDirtyNotesInNotebook:(ENNotebookMO *)notebook;
- (NSArray *) noteGUIDsToExpunge;
- (NSUInteger) numberOfActiveNotes;
- (NSUInteger) numberOfActivePersonalNotes;
- (BOOL) doesActiveNoteCountExceed:(NSUInteger)minimum;
- (BOOL) doesCountOfNotesMatchingPredicate:(NSPredicate *)predicate
                                    exceed:(NSUInteger)minimum;

- (NSUInteger) numberOfInactivePersonalNotes;
- (NSUInteger) numberOfActiveBusinessNotes;
- (NSUInteger) numberOfInactiveBusinessNotes;
- (NSUInteger) numberOfActiveNotesWithTag:(ENTagMO *)tag;
- (NSUInteger) numberOfActiveNotesInNotebook:(ENNotebookMO *)notebook;
- (NSUInteger) numberOfActiveLocalNotes; // refers to notes in local notebooks

- (ENNoteMO *) noteWithLocalUUID:(NSString *)localUUID;
- (ENNoteMO *) noteWithGUID:(NSString *)guid;
- (NSArray *) notesWithGUIDs:(NSArray *)guids;
- (NSArray *) notesWithObjectIDs:(NSArray *)objectIDs;
- (NSArray *) notesMatchingPredicate:(NSPredicate *)predicate;
- (NSArray *) notesMatchingPredicate:(NSPredicate *)predicate sortedBy:(NSArray *)sortDescriptors;
- (NSArray *) noteIDsMatchingPredicate:(NSPredicate *)predicate;
- (NSArray *) noteIDsMatchingPredicate:(NSPredicate *)predicate sortedBy:(NSArray *)sortDescriptor;
- (NSArray *) resourcesMatchingPredicate:(NSPredicate *)predicate;
- (NSArray *) resourceIDsMatchingPredicate:(NSPredicate *)predicate;
- (NSArray *) inactivePersonalNotes;
- (NSArray *) distinctNoteCitiesMatchingPredicate:(NSPredicate *)predicate;
- (NSArray *) distinctNoteCities;

- (NSArray *) uncachedNotesMatchingPredicate:(NSPredicate *)predicate
                                    sortedBy:(NSArray *)sortDescriptors
                                  fetchLimit:(NSUInteger)limit;

/**
 * Existing note with the specified coredata id
 */
- (ENNoteMO *) noteWithLocalID: (NSManagedObjectID *) moid;


// If a note, or any of it's resources, contain the specified cacheStatus,
// they'll be returned here...
- (NSArray *) notesWithCachingStatus:(ENManagedObjectCachingStatus)cachingStatus;

- (uint64_t) totalSizeOfNotesMatchingPredicate:(NSPredicate *)predicate;
- (uint64_t) cacheSizeOfNotesMatchingPredicate:(NSPredicate *)predicate;

- (NSArray *) notesWithCachedData;
- (uint32_t) numberOfNotesWithCachedData;

- (ENNoteMO *) createNote;
- (void) deleteNote:(ENNoteMO *)note;
- (void) deleteNotes:(NSArray *)notes;
- (void) unDeleteNote:(ENNoteMO *)note;
- (void) unDeleteNotes:(NSArray *)notes;

- (NSDictionary *) mapOfNotesWithGUIDs:(NSArray *)guids;

- (uint64_t) sizeOfNotesMatchingPredicate:(NSPredicate *)predicate;

- (BOOL) validateNoteTitle:(id*)value
                     error:(NSError *__autoreleasing *)error;
- (BOOL) validateNoteSourceURL:(id*)value
                         error:(NSError *__autoreleasing *)error;

- (id) newNotesControllerWithPredicate:(NSPredicate *)predicate
                       sortDescriptors:(NSArray *)sortDescriptors;
- (id) newNotesControllerWithPredicate:(NSPredicate *)predicate
                       sortDescriptors:(NSArray *)sortDescriptors
                    sectionNameKeyPath:(NSString*)sectionName;
- (id) newNotesControllerOfClass:(Class)controllerClass
                   withPredicate:(NSPredicate *)predicate
                 sortDescriptors:(NSArray *)sortDescriptors
              sectionNameKeyPath:(NSString*)sectionName;

- (BOOL) setENMLContent:(NSString *)contents
                forNote:(ENNoteMO *)note
                  error:(NSError *__autoreleasing *)error;

- (void) invalidateContentForNote:(ENNoteMO *)note;
- (void) noteChangedDirtyValue:(ENNoteMO *)note;

- (ENNoteMO *)duplicateNote:(ENNoteMO *)sourceNote;
- (ENNoteMO *)duplicateNote:(ENNoteMO *)sourceNote toNotebook:(ENNotebookMO *)notebook;
- (BOOL) canEasilyMoveNote:(ENNoteMO *)note
              fromNotebook:(ENNotebookMO *)oldNotebook
                toNotebook:(ENNotebookMO *)newNotebook;
- (void) moveNote:(ENNoteMO *)note toNotebook:(ENNotebookMO *)notebook;

- (void) addTagsWithNames:(NSSet *)tagNames toNote:(ENNoteMO *)note;

#pragma mark -
#pragma mark Notebooks
- (NSArray *) dirtyNotebookIDs;
- (NSArray *) notebookGUIDsToExpunge;

- (NSUInteger) numberOfNotebooks;
- (NSUInteger) numberOfPersonalNotebooks;
- (NSArray *) allNotebooks;
- (NSArray *) allPersonalNotebooks;
- (NSArray *) allBusinessNotebooks;
- (NSArray *) notebooksWithGUIDs:(NSArray *)guids;
- (ENNotebookMO *) notebookWithGUID:(NSString *)guid;
- (ENNotebookMO *) notebookWithName:(NSString *)name;

- (NSArray *) notebooksWithNamePrefix:(NSString *)name;
- (NSArray *) notebooksWithNamePrefix:(NSString *)name
                          inAccountID:(id)accountID;

- (NSArray *) notebooksInStackNamed:(NSString *)stackName;

/**
 * Existing notebook with the specified coredata id
 */
- (ENNotebookMO *) notebookWithLocalID: (NSManagedObjectID *) moid;

/**
 * Returns latest dateUpdated of the notes in the notebook
 */
- (NSDate *) latestDateUpdatedOfNotesInNotebook:(ENNotebookMO *)notebook;

- (ENNotebookMO *) defaultNotebook;

- (ENNotebookMO *) createNotebook;
- (ENNotebookMO *) createNotebookWithName:(NSString *)name;
- (ENNotebookMO *) createBusinessNotebookWithName:(NSString *)name;
- (ENPublishingMO *) createPublishingForNotebook:(ENNotebookMO *)notebook;
- (ENBusinessNotebookMO *) createBusinessNotebookDescription;

- (void) deleteNotebook:(ENNotebookMO *)notebook;
- (void) deleteNotebooks:(NSArray *)notebooks;
- (void) deleteNotebookStackNamed:(NSString *)stackName;
- (void) deletePublishing:(ENPublishingMO *)publishing;
- (void) renameNotebookStackFrom:(NSString *)oldStackName
                              to:(NSString *)newStackName;

- (NSDictionary *) mapOfNotebooksWithGUIDs:(NSArray *)guids;

- (BOOL) validateName:(id*)value
          forNotebook:(ENNotebookMO *)notebook
                error:(NSError *__autoreleasing *)error;

- (BOOL) validateNotebookStackName:(id *)value
                             error:(NSError *__autoreleasing *)error;

- (BOOL) validateURI:(id*)value
          forNotebook:(ENNotebookMO *)notebook
                error:(NSError *__autoreleasing *)error;

- (NSString *) sanitizedAndUniquifiedNotebookNameFromName:(NSString *)name;

- (id) newPersonalNotebooksControllerWithPredicate:(NSPredicate *)predicate
                                   sortDescriptors:(NSArray *)sortDescriptors;

- (id) newNotebooksControllerWithPredicate:(NSPredicate *)predicate
                           sortDescriptors:(NSArray *)sortDescriptors;

#pragma mark -
#pragma mark Resources
- (NSArray *) dirtyResourceIDs;
- (NSArray *) resourceGUIDsToExpunge;

- (NSArray *) resourcesWithMimeTypes:(NSArray *)mimeTypes;
- (NSArray *) resourcesWithGUIDs:(NSArray *)guids;
- (ENResourceMO *) resourceWithGUID:(NSString *)guid;
- (ENResourceMO *) resourceWithLocalID: (NSManagedObjectID *)moid;
- (ENResourceMO *) createResource;
- (void) populateAttributesOnResource:(ENResourceMO *)resource;
- (ENResourceMO *) findOrCreateResourceInNote:(ENNoteMO *)note
                                     withData:(NSData *)data
                                     mimeType:(NSString *)mimeType;
- (ENResourceMO *) findOrCreateResourceInNote:(ENNoteMO *)note
                     withContentsOfItemAtPath:(NSString *)sourcePath 
                                     filename:(NSString *)filename
                             removingWhenDone:(BOOL)removeWhenDone;

- (void) deleteResource:(ENResourceMO *)resource;
- (void) deleteResources:(NSArray *)resources;
- (void) prepareResourceForReuse:(ENResourceMO *)resource;

- (uint64_t) sizeOfResourcesMatchingPredicate:(NSPredicate *)predicate;

- (id) newResourcesControllerWithPredicate:(NSPredicate *)predicate
                           sortDescriptors:(NSArray *)sortDescriptors;

#pragma mark -
#pragma mark Saved Searches
- (NSArray *) allSavedSearches;
- (NSArray *) dirtySavedSearchIDs;
- (NSArray *) savedSearchGUIDsToExpunge;

- (NSArray *) savedSearchesWithGUIDs:(NSArray *)guids;
- (ENSavedSearchMO *) savedSearchWithGUID:(NSString *)guid;
- (ENSavedSearchMO *) savedSearchWithName:(NSString *)name;

- (NSDictionary *) mapOfSearchesWithGUIDs:(NSArray *)guids;


- (ENSavedSearchMO *) createSavedSearch;
- (ENSavedSearchMO *) createSavedSearchWithName:(NSString *)name;
- (void) deleteSavedSearch:(ENSavedSearchMO *)savedSearch;
- (void) deleteSavedSearches:(NSArray *)savedSearches;

- (BOOL) validateName:(id*)value
       forSavedSearch:(ENSavedSearchMO *)savedSearch
                error:(NSError *__autoreleasing *)error;
- (NSString *) sanitizeAndUniquifySavedSearchName:(NSString *)name;

- (id) newSavedSearchesControllerWithPredicate:(NSPredicate *)predicate
                               sortDescriptors:(NSArray *)sortDescriptors;

#pragma mark -
#pragma mark Shared notebooks
- (NSArray *) dirtySharedNotebookIDs;
- (NSArray *) sharedNotebookIDsToExpunge;
- (NSArray *) sharedNotebooksMatchingPredicate:(NSPredicate *)predicate;
- (NSArray *) sharedNotebooksWithIdsIn:(NSArray *)sharedNotebookIds;

- (void) deleteSharedNotebook:(ENSharedNotebookMO *)sharedNotebook;
- (ENSharedNotebookMO *) createSharedNotebook;

#pragma mark -
#pragma mark Tags
- (NSArray *) allTags;
- (NSArray *) dirtyTagIDs;
- (NSArray *) tagGUIDsToExpunge;

- (NSArray *) tagsWithGUIDs:(NSArray *)guids;
- (NSArray *) tagsWithParentGUID:(NSString *)guid;
- (ENTagMO *) tagWithGUID:(NSString *)guid;
- (ENTagMO *) tagWithGUID:(NSString *)guid
              inAccountID:(id)accountID;
- (ENTagMO *) tagWithName:(NSString *)name;
- (ENTagMO *) tagWithName:(NSString *)name
              inAccountID:(id)accountID;

- (NSArray *) tagsWithNames:(NSArray *)names
                inAccountID:(id)accountID;

- (NSArray *) tagsWithNamePrefix:(NSString *)name;
- (NSArray *) tagsWithNamePrefix:(NSString *)name
                     inAccountID:(id)accountID;

- (NSUInteger) numberOfTags;
- (NSUInteger) numberOfPersonalTags;
- (NSUInteger) numberOfBusinessTags;
- (NSUInteger) numberOfTagsMatchingPredicate:(NSPredicate *)predicate;

- (NSArray *) distinctTagNames;
- (NSArray *) distinctTagNamesWithNoteCounts;

- (ENTagMO *) createTagInAccountID:(id)accountID;
- (ENTagMO *) createTagWithName:(NSString *)name
                    inAccountID:(id)accountID;
- (ENTagMO *) findOrCreateTagWithName:(NSString *)name
                          inAccountID:(id)accountID;

- (void) deleteTag:(ENTagMO *)tag;
- (void) deleteTags:(NSArray *)tags;

- (NSDictionary *) mapOfTagsWithGUIDs:(NSArray *)guids;

- (BOOL) validateName:(id*)value
               forTag:(ENTagMO *)tag
                error:(NSError *__autoreleasing *)error;
- (NSString *) sanitizeAndUniquifyTagName:(NSString *)name
                              inAccountID:(id)accountID;

- (id) newPersonalTagsControllerWithPredicate:(NSPredicate *)predicate
                              sortDescriptors:(NSArray *)sortDescriptors
                              sectionsEnabled:(BOOL)sectionsEnabled;

- (id) newBusinessTagsControllerWithPredicate:(NSPredicate *)predicate
                              sortDescriptors:(NSArray *)sortDescriptors
                              sectionsEnabled:(BOOL)sectionsEnabled;

- (id) newAllTagsControllerWithPredicate:(NSPredicate *)predicate
                         sortDescriptors:(NSArray *)sortDescriptors
                         sectionsEnabled:(BOOL)sectionsEnabled;

- (id) newLinkedTagsControllerWithAccountID:(id)accountID
                                  predicate:(NSPredicate *)predicate
                            sortDescriptors:(NSArray *)sortDescriptors
                            sectionsEnabled:(BOOL)sectionsEnabled;


#pragma mark -
#pragma mark Users
- (NSArray *) allUserProfiles;
- (NSArray *) userProfilesWithIdsIn:(NSArray *)ids;
- (id<ENUserProfile>) userProfileWithId:(int32_t)userId;
- (id<ENUserProfile>) createUserProfile;


/**
 * Convenience property to get the root persistent store's unique identifier
 */
@property (readonly,nonatomic) NSString * persistentStoreIdentifier;


/**
 * Empty ENML used to pre-fill a new note
 */
@property (readonly,nonatomic) NSString * emptyENML;


/**
 * idle saves will not reset the managed object context more frequently than this interval
 */
@property NSTimeInterval idleSaveResetInterval;


/**
 * A special notebook which should be hidden from the user which contains all shared notes for 
 * which we do not have access to the notebook.
 */
@property (readonly,nonatomic) ENNotebookMO * danglingSharedNotesNotebook;


/**
 *
 * ===================================================================================================
 * YOU SHOULDN'T NEED TO USE THE FOLLOWING METHODS UNLESS YOU ARE SUBCLASSING OR DOING SOMETHING ELITE
 * ===================================================================================================
 *
 */
- (id) fetchEntityOfType:(NSString *)entityName withPredicate:(NSPredicate *)predicate;
- (NSArray *) fetchEntitiesOfType:(NSString *)entityName withPredicate:(NSPredicate *)predicate;
- (NSManagedObject *) createEntityOfName:(NSString *)entityName;


@end

