#ifndef ENNoteMO_HEADER_GUARD
#define ENNoteMO_HEADER_GUARD

#import "_ENNoteMO.h"
#import <CoreLocation/CoreLocation.h>

@class ENNotebookMO;
@class ENResourceMO;
@class ENTagMO;
@class EDAMNote;
@class EDAMNoteRestrictions;

extern NSString * const ENNoteLocationUnknown;

extern NSString * const ENNoteSourceApplicationBusinessCards;
extern NSString * const ENNoteSourceApplicationMoleskine;
extern NSString * const ENNoteSourceApplicationPostIt;


@interface ENNoteMO : _ENNoteMO {}

@property (copy, nonatomic) CLLocation *location;
@property (nonatomic) int32_t creatorId;
@property (nonatomic) int32_t lastEditorId;

/**
 * A set of EDAMSharedNote structures belonging to this note, which convey single-note-sharing info.
 * Passive-aggressive comment: i don't like this property name, but its what the service uses.
 */
@property (nonatomic,strong) NSSet * sharedNotes;

/**
 * If this user is seeing this note via a single-note-share with no notebook, the restrictions
 * on what can and cannot be done are here.
 */
@property (nonatomic,strong) EDAMNoteRestrictions * restrictions;


/**
 * For a shared note w/o a notebook, the note store URL for fetching this note and its data from the service
 */
@property (nonatomic,strong) NSURL * noteStoreURL;


/**
 * The shardID that the note lives on.
 * For dangling shared notes, this will be the shardID of whereever that note lives.
 * For notes in linked notebooks, this will be the shardID of the linked notebook.
 * For notes in personal notebooks, this will be the shardID of the user's account.
 */
@property (nonatomic,strong) NSString * shardID;


+ (void) registerENMLTransformerClass:(Class)transformerClass;
+ (Class) enmlTransformerClass;

+ (NSPredicate *) predicateForNotesInTrash;
+ (NSPredicate *) predicateForNotesInStack:(NSString *)stack;
+ (NSPredicate *) predicateForNotesInNotebook:(ENNotebookMO *)notebook;
+ (NSPredicate *) predicateForNotesWithTag:(ENTagMO *)tag;
+ (NSPredicate *) predicateForNotesWithGeolocation;
+ (NSPredicate *) predicateForNotesInRegion:(CLRegion *)region;

- (ENResourceMO *) resourceWithGUID:(NSString *)guid;
- (ENResourceMO *) resourceWithHash: (NSData *) hash;

- (NSString *) backingFileExtension;
- (NSString *) absolutePathForBackingFile;
- (NSArray *) allBackingFilePaths;
@property (readonly,nonatomic) NSString * absolutePathForBackingDirectory;

// Returns the total size of note, including resources
- (uint32_t) totalNoteSize;

// Returns the cached size of the note, including resources
- (uint32_t) cachedNoteSize;

// Returns the cache status of the note and it's resources
// ENManagedObjectCachingStatusComplete means that everything
// exists.  
- (ENManagedObjectCachingStatus) totalCacheStatus;

- (void) invalidateLocalContent;
- (BOOL) needsSearchContent;

/**
 * Convenience property
 */
@property (readonly,nonatomic) BOOL isBusinessNote;

#pragma mark - Note sync callbacks

/**
 *  This callback will be invoked before this note instance is updated with new server values.
 *
 *  @param serverNote the server note data that was just synced from the server
 */
- (void) willUpdateFromServerNote:(EDAMNote *)serverNote;

/**
 *  This callback will be invoked after this note instance was updated with new server values.
 *
 *  @param serverNote the server note data that was just synced from the server
 */
- (void) didFinishUpdatingFromServiceNote:(EDAMNote *)serverNote;

/**
 *  This callback will be invoked before the serverNote instance is updated with new values from this instance.
 *
 *  @param serverNote the server note data that is about to be synced with server
 */
- (void) willBeginPreparingNoteForUpload:(EDAMNote *)serverNote;

/**
 *  This callback will be invoked after the serverNote instance was updated with new values from this instance.
 *
 *  @param serverNote the server note data that is about to be synced with server
 */
- (void) didFinishPreparingNoteForUpload:(EDAMNote *)serverNote;

#pragma mark - Server download progress

// you need to enable track download progress before the API below will have any affect
+ (void) setTrackDownloadProgress:(BOOL)enableProgress;
+ (BOOL) trackDownloadProgress;

// compute the server payload left to download, which can be less than total note size
- (uint32_t) totalDownloadSize;

// prime and reset for the next download session; must be on main thread!
- (void) resetDownloadProgress:(uint32_t)totalDownloadSize;

// true iff we're currently downloading note content from the server; must be on main thread!
@property (nonatomic) BOOL isLoading;
// percent loaded in this session; must be on main thread!
@property (nonatomic) uint32_t percentLoaded;
// bytes downloaded already in this session; must be on main thread!
@property (nonatomic) uint32_t bytesDownloaded;
// total bytes we expect to download this session; must be on main thread!
@property (nonatomic) uint32_t downloadSize;

#pragma mark

@property (copy, nonatomic) NSDictionary *lastKnownServerValues;


/**
 * Handy for logging
 */
@property (readonly,nonatomic) NSString * shortDescription;


/**
 * This is a shared note and we don't have access to the notebook
 */
@property (readonly,nonatomic) BOOL isDanglingSharedNote;


/**
 *  Predicate for filtering out any dangling shared notes.  "Dangling" shared notes
 *  are notes shared with us where we don't have access to the notebook.
 *
 *  @return a predicate, duh.
 */
+ (NSPredicate *) noDanglingSharedNotesPredicate;


@end


/**
 *  optional methods on ENNoteMO that will be used if available
 */
@interface ENNoteMO(OptionalMethods)
- (NSString *) absolutePathForRawServerFile;
- (NSString *) enmlContent;
- (void) setEnmlContent:(NSString *)enmlContent;
- (void) setSearchText:(NSString *) searchText;
- (void)decodeApplicationDataMap:(NSDictionary *)map;
- (void)encodeApplicationDataMap:(NSMutableDictionary *)map;
@property NSDictionary *classifications;
@end


#endif
