//
//  ENSyncEngine.h
//  Evernote
//
//  Created by Steve White on 4/16/09.
//  Copyright 2009 Evernote Corporation. All rights reserved.
//

#ifndef ENSyncEngine_HEADER_GUARD
#define ENSyncEngine_HEADER_GUARD

#import <Foundation/Foundation.h>

#import "EDAMClientCache.h"
#import "EDAMHTTPClient.h"
#import "ENCoreDataLocalStore.h"
#import "ENSyncState.h"

#import "NSDateEDAMAdditions.h"

#if defined INTERNAL_BUILD && INTERNAL_BUILD != 0
#   define SyncStartTime() NSDate * saveStart = [NSDate date]
#   define SyncEndTime() SyncLogDebug(@"processing time: %f", [saveStart timeIntervalSinceNow] * -1.0)
#else
#   define SyncStartTime() { }
#   define SyncEndTime() { }
#endif

#define SyncLogDebug(...)  AppLogDebug(__VA_ARGS__)
#define SyncLogInfo(...)  AppLogInfo(__VA_ARGS__)
#define SyncLogWarn(...)  AppLogWarn(__VA_ARGS__)
#define SyncLogError(...) AppLogError(__VA_ARGS__)

extern NSString * const ENSyncEngineWWANSyncDisabledPreferenceKey;

extern NSString * const ENSyncEngineErrorDomain;

extern NSString * const ENSyncEngineDidStartNotification;
extern NSString * const ENSyncEngineDidStopNotification;
extern NSString * const ENSyncEngineWillUploadLocalChangesNotification;
extern NSString * const ENSyncEngineWillUploadEntityNotification;
extern NSString * const ENSyncEngineDidUploadEntityNotification;
extern NSString * const ENSyncEngineDidFailWithErrorsNotification;
extern NSString * const ENSyncEngineDidChangeStatusMessageNotification;

extern NSString * const ENSyncEngineDidChangeNoteDataHashNotification;

extern NSString * const ENSyncEngineDidUploadEntityNotification_SuccessKey;
extern NSString * const ENSyncEngineDidUploadEntityNotification_USNKey;


extern int const ENSyncEngineChunkSize;

enum {
  ENSyncEngine_MissingCredentials = -1000,
  ENSyncEngine_InvalidUsername = -1001,
  ENSyncEngine_InvalidPassword = -1002,
  ENSyncEngine_EDAMVersionCheckFailed = -1003,
  ENSyncEngine_ExceptionRaised = -1004,
  ENSyncEngine_CommFailure = -1005,
  ENSyncEngine_InvalidAuthToken = -1006,
  ENSyncEngine_ExpiredPassword = -1100,
  ENSyncEngineCRUDError = -2000,
  ENSyncEngineLocalStoreSaveException = -3000,
};

typedef enum {
  ENSyncEngineConflictResolutionMerge,
  ENSyncEngineConflictResolutionServerWins,
  ENSyncEngineConflictResolutionClientWins,
  ENSyncEngineConflictResolutionDuplicate,
  ENSyncEngineConflictResolutionSkip = -1,
} ENSyncEngineConflictResolutionType;

typedef struct {
  unsigned int noteApplicationDataDecoder:1;
  unsigned int noteApplicationDataEncoder:1;
  unsigned int resourceApplicationDataDecoder:1;
  unsigned int resourceApplicationDataEncoder:1;
  unsigned int resourceWantsResourceAlternateData:1;
} ENSyncEngineEntityFlagsType;

extern ENSyncEngineEntityFlagsType gENSyncEngineEntityFlags;

@class EDAMUsage;
@class ENSyncOperation;
@protocol ENSyncEngineDelegate;
@protocol ENSyncEngineConflictDelegate;

@interface ENSyncEngine : NSObject {
#if TARGET_OS_IPHONE
  UIBackgroundTaskIdentifier _backgroundTaskID;
#endif
  
  ENCoreDataLocalStore *_mainLocalStore;
  ENCoreDataLocalStore *_filteredLocalStore;
  EDAMHTTPClient * _edamClient;
  EDAMHTTPClient * _edamNoteStoreClient;
  EDAMClientCache *_edamClientCache;
  ENSyncState *_currentSyncState;
  ENSyncState *_mainSyncState;
  EDAMUsage *_edamUsage;

  BOOL _linkedAccount;
  EDAMLinkedNotebook *_edamLinkedNotebook;
    
  __weak id<ENSyncEngineDelegate> _delegate;
  __weak id<ENSyncEngineConflictDelegate> _conflictDelegate;
  struct {
		unsigned int delegateDidChangeStatusMessage:1;
	} _delegateFlags;
  
  NSCondition *_downloadedChunksLock;
	NSMutableArray *_downloadedChunks;
  
  BOOL _syncShouldCancel;
  BOOL _isPerformingFullSync;
  BOOL _needsUserProfileRefresh;
  
  NSMutableArray *_syncErrors;
  void (^_syncFinishedCallback)(ENSyncEngine *);
}

@property (strong, nonatomic) ENSyncState *mainSyncState;

@property (strong, nonatomic) ENCoreDataLocalStore *mainLocalStore;
@property (strong, nonatomic) ENCoreDataLocalStore *filteredLocalStore;

@property (strong, nonatomic) EDAMHTTPClient *edamClient;
@property (strong, nonatomic) EDAMClientCache *edamClientCache;

@property (strong, nonatomic) EDAMUsage *edamUsage;
@property (strong, nonatomic) NSString *noteContentClassFilter;

@property (strong, nonatomic) NSString *cacheDirectory;

@property (strong, readonly, nonatomic) NSString *statusMessage;
@property (readonly, nonatomic) NSArray *syncErrors;
@property (assign, nonatomic) BOOL suppressNetworkErrors;

@property (readonly, nonatomic, getter=isExecuting) BOOL executing;

@property (weak, nonatomic) id<ENSyncEngineDelegate> delegate;
@property (weak, nonatomic) id<ENSyncEngineConflictDelegate> conflictDelegate;

@property (readonly, nonatomic) id uploadingEntityID;

- (void) cancel;

- (BOOL) sync;
- (BOOL) syncForcingAuthRefresh:(BOOL)authRefresh;
- (BOOL) syncIgnoringNetworkPreference;
- (BOOL) syncLocalChangesWithIds:(NSArray *)localChangeIds;
- (BOOL) syncIgnoringNetworkPreference:(BOOL)ignoringNetworkPreference
                    forcingAuthRefresh:(BOOL)forceAuthRefresh;

- (void) refreshUserInfo;

@property (readonly, nonatomic) NSArray *syncOperations;
- (void) addSyncOperation:(ENSyncOperation *)syncOperation;
- (void) removeSyncOperation:(ENSyncOperation *)syncOperation;

@end

@protocol ENSyncEngineConflictDelegate<NSObject>
@optional
// The receiver is invoked when there is any change between the notes
// (could be a metadata change, e.g. author, or it could be a content change)
// If implemented, this is the only delegate method that will be invoked during
// conflict resolution.
- (ENSyncEngineConflictResolutionType) resolveConflictBetweenLocalNoteWithID:(id)localNoteID
                                                               andServerNote:(id)serverNote
                                                              withServerENML:(id)serverENML;

// The sync engine is responsible for managing metadata conflicts.  This method
// will only be invoked if the content has changed between the server & client.
- (ENSyncEngineConflictResolutionType) resolveContentConflictBetweenLocalNoteWithID:(id)localNoteID
                                                                      andServerNote:(id)serverNote
                                                                     withServerENML:(id)serverENML;


// The only acceptable resolution types are client wins, or server wins.
// Duplicate will be treated as a server wins
- (ENSyncEngineConflictResolutionType) resolveConflictBetweenLocalResourceWithID:(id)localResourceID
                                                               andServerResource:(id)serverResource;


- (ENSyncEngineConflictResolutionType) resolveConflictBetweenLocalClientPreference:(id)localClientPreference
                                                                    andServerValue:(NSArray *)serverValue;

- (BOOL) resolvePermissionErrorForNoteWithID:(id)localNoteID
                       withResolutionMessage:(NSString **)resolutionMessage
                                inSyncEngine:(ENSyncEngine *)syncEngine;


/**
 * A notebook has been expunged on the service but we have dirty notes locally.
 * Decide where to put the dirty notes.  Called from within a read/write store block
 */
- (ENNotebookMO *) resolveDestinationForDirtyNotesFromExpungedNotebook: (ENNotebookMO *) notebook
                                                            localStore: (ENCoreDataLocalStore*) localStore;


@end

@interface NSError (ENSyncEngineErrorHelper)
- (BOOL) isFatalEDAMError;
@end

#import "ENSyncEngineDelegate.h"

#endif
