//
//  NSStringAdditions.h
//  Evernote
//
//  Created by Steve White on 10/5/07.
//  Copyright 2007 Evernote Corporation. All rights reserved.
//

#ifndef ENSharedUtilities_NSStringAdditions
#define ENSharedUtilities_NSStringAdditions

#if TARGET_OS_IPHONE
#import <Foundation/Foundation.h>
#else
#import <Cocoa/Cocoa.h> // For NSAttributedString
#endif

@interface NSString (ENSharedUtilities)

- (NSString *)trimWhitespace;

/**
 * Create a unique string
 */
+ (NSString *) uniqueString;


/**
 * DOCME
 */
- (NSString *) xhtmlEncoded;


/**
 * Returns true if this string starts and ends with "
 */
- (BOOL) isQuoted;


/**
 * Strip beginging and ending double quotes if they are present.
 */
- (NSString *) stripQuotes;


/**
 * Return the string surrounded by double quotes if it contains any whitespace
 */
- (NSString *) quoteIfContainsWhitespace;


/**
 * Return the string with all sequences of characters from the given character set replaced by
 * the replacement string.
 * If max is greater than zero, the result string will not be longer than max characters.
 */
- (NSString *) stringByCollapsingSequencesOf: (NSCharacterSet *) charset
                                        with: (NSString *) replacement
                             maxResultLength: (NSUInteger) max;


/**
 * Escape any occurences of '*' and '?' with backslashes.
 */
- (NSString *) stringByEscapingLIKEWildcardCharacters;


/**
 * Returns a string with invalid XML characters removed
 */
- (NSString *)EN_validXMLString;

/**
 * turn a string into something that can used as a legal file name
 * DO NOT USE FOR ENTIRE PATHS
 */
- (NSString *) sanitizeFilenameString;

/**
 * Write the string to an output stream using the specified encoding.
 */
- (NSInteger) writeToStream: (NSOutputStream *) outStream usingEncoding: (NSStringEncoding) encoding  error: (NSError**) error;


/**
 * Compare receiver to another string, but consider null equal to the empty string
 */
- (BOOL) isEqualToStringWithEmptyEqualToNull: (NSString *) s;


/**
 * Return the string with all control characters removed
 */
- (NSString *) stringWithoutControlCharacters;


#if !TARGET_OS_IPHONE
/**
 * Create an attributed string from this string that is a link to the destination
 */
- (NSAttributedString *) attributedStringWithLinkTo: (NSString *) destination;

/**
 * Create an attributed string from this string that is a link to the destination.
 *
 * @param destination the destination URL as a string
 * @param linkColor the color to make the link display.  Pass nil for standard blue
 * @param underlineIt YES if the link should be underlined
 *
 * @return The new attributed string
 */

- (NSAttributedString *) attributedStringWithLinkTo: (NSString *) destination
                                              color: (NSColor *) linkColor
                                          underline: (BOOL) underlineIt;


/**
 * Create an attributed string from this string that looks like a link (with underline)
 */
- (NSAttributedString *) blueUnderlinedAttributedString;


/**
 * Create an attributed string from this string that looks like a link (no underline)
 */
- (NSAttributedString *) blueAttributedString;


#endif


/**
 * Split the receiver into tokens and split any CJK tokens up into individual characters,
 * and append each one to the supplied mutable string, preceded by 'separatorString',
 * filtering out any stop words.
 */
- (void) tokenizeAndSplitCJKTokensAppendingToString: (NSMutableString *) wordsString
                                        separatedBy: (NSString *) separatorString
                                      appendToArray: (NSMutableArray *) wordsArray
                                          stopWords: (NSSet *) stopWords;


/**
 * Split the receiver into tokens and split any CJK tokens up into individual characters,
 * and append each one to the supplied mutable string, preceded by 'separatorString',
 * filtering out any stop words.
 */
- (void) tokenizeAndSplitCJKTokensAppendingToString: (NSMutableString *) wordsString
                                        separatedBy: (NSString *) separatorString
                                          stopWords: (NSSet *) stopWords;


/**
 * Split the receiver into tokens and split any CJK tokens up into individual characters,
 * and append each one to the supplied mutable string, preceded by 'separatorString'
 */
- (void) tokenizeAndSplitCJKTokensAppendingToString: (NSMutableString *) wordsString
                                        separatedBy: (NSString *) separatorString;

/**
 * Split the receiver into tokens, even if the receiver is a CJK string,
 * and append each one to the supplied mutable string, preceded by 'separatorString',
 * filtering out any stop words.
 */
- (void) tokenizeAndAppendToString: (NSMutableString *) wordsString
                       separatedBy: (NSString *) separatorString
                         stopWords: (NSSet *) stopWords;

/**
 * Return the string with all diacritics removed, and converted to lowercase.
 */
- (NSString *)stringByLowercasingAndStrippingDiacriticals;

/**
 * Return the original string, unless it is nil, in which case, return the empty
 * string @""
 */
+ (NSString *) stringOrEmptyString: (NSString *) s;


/**
 * Grab the substring between the first sequential pair of strings
 */
- (NSString *) substringBetween: (NSString *) start and: (NSString *) end;


/**
 * Copy the string to the clipboard as a URL.
 */
- (void)writeToPasteboardAsURL;


/**
 * This behaves differently than stringByAddingPercentEscapesUsingEncoding.
 */
- (NSString *) wwwFormURLEncode;


/**
 * Truncate a string to a specified number of composed character sequences and append an ellipsis if necessary.
 * Also strips leading/trailing whitespace and returns "-" if the string is empty.
 */
- (NSString *)stringByTruncatingToMaxNumberOfComposedCharacterSequences:(NSUInteger)maxLength;


/**
 * Strips out an <email address> from a person name if one is present,
 * returns the string unmodified if one is not present.
 */
- (NSString *)stringByRemovingEmailAddress;

/* Determines if this string starts with a capital letter */
-(BOOL) startsWithCaps;

/* Repeats a string str, n number of times using the delimiter. 
 * e.g. str = apple, number of times = 4, delimiter = #
 * Final string = apple#apple#apple#apple
 * Always returns a new string which is autoreleased
 */
-(NSString*) stringByRepeatingNTimes:(NSUInteger) numberOfRepeats usingDelimter:(NSString*) delimiter;

/**
 * Removes leading whitespace from this string
 */
-(NSString*) removeLeadingWhitespace;

/**
 * Returns a new autoreleased string with NormalizationKD form
 */
-(NSString *)stringWithNormalizationKD;

/**
 * Checks if string contains a scheme, if it doesn't it uses the default scheme
 * @param defaultScheme   Default scheme to be used, i.e.: http
 */
- (NSString *)stringByAddingURLSchemeIfNeeded:(NSString *)defaultScheme;

/**
 * Checks if this string is a number representation
 * A string is a number if it starts with a digit
 * Cannot use the inbuilt integerValue method because it returns 0 if not a number, 
 * and wont catch different locale numbers. This uses the NSCharacterSet.decimalDigits
 * to check for a number representation
 * Based on the apple documentation, NSString.integervalue uses formatting information to 
 * for non localized strings. This seems like a more robust check.
 */
- (BOOL) hasPositiveIntegerPrefixWithoutFormatting;

/**
 * Returns a base64 encoded string
 */
- (NSString*) base64Encode;

/**
 * Decodes this base64 encoded string
 */
- (NSString*) base64Decode;

/**
 * @return YES if a word matches the prefixString
 * Example 'ear' should match: 'Early Candidates' and 'Big Ears' but not 'Searched Stuff'
 */
- (BOOL)matchesWordWithPrefixString:(NSString *)prefixString;


/**
 * @return YES if the prefixString is matched (case/diacritical/width-insensitive) at the
 * beginning of a word boundary in the receiver.
 * Example 'ear' should match: 'Early Candidates' and 'Big Ears' but not 'Searched Stuff'
 * Example 'v4 ux' should match 'Mac V4 UX'
 */
- (BOOL)matchesAtWordBoundaryWithPrefixString:(NSString *)prefixString;

/**
 * @return YES if the token is a CJK token
 */
- (BOOL)isTokenCJK:(CFStringTokenizerTokenType)tokenType range:(CFRange)tokenRange;

/**
 * @return YES if the token is a CJ token
 */
- (BOOL)isTokenCJ:(CFStringTokenizerTokenType)tokenType range:(CFRange)tokenRange;

/**
 * @return YES if any char is a CJK
 *
 * Note, this method is slightly different from isTokenCJK::, which only checks whether the first char is CJK or not.
 */
- (BOOL)isAnyCharCJK;

/**
 * Compares the receiver's onscreen pixel width using the default font with that of otherString.
 * Returns NSOrderedAscending if the receiver's width is less than that of otherString.
 * Returns NSOrderedDescending if the receiver's width is greater than that of otherString.
 * Returns NSOrderedSame if the receiver's width is equal to that of otherString.
 * Results may not be accurate if either string contains newlines.
 */
- (NSComparisonResult)compareOnScreenWidth:(NSString *)otherString;


/** 
 * @return The portion of self, assumed to be a URL, before the fragment delimiter (#).
 */
- (NSString *)stringByStrippingURLFragment;

/**
 *  Takes the data of an ENWebResource or WebResource object and converts it
 *  to a string with a certain encoding.
 */
+ (NSString *)stringFromData:(NSData *)data textEncodingName:(NSString *)textEncoding;

@end

/**
 * For the analytics code (ENAnalytics.h/m)
 */
@interface NSString (GTMNSStringURLArgumentsAdditions)

/// Returns a string that is escaped properly to be a URL argument.
//
/// This differs from stringByAddingPercentEscapesUsingEncoding: in that it
/// will escape all the reserved characters (per RFC 3986
/// <http://www.ietf.org/rfc/rfc3986.txt>) which
/// stringByAddingPercentEscapesUsingEncoding would leave.
///
/// This will also escape '%', so this should not be used on a string that has
/// already been escaped unless double-escaping is the desired result.
- (NSString*)gtm_stringByEscapingForURLArgument;

/// Returns the unescaped version of a URL argument
//
/// This has the same behavior as stringByReplacingPercentEscapesUsingEncoding:,
/// except that it will also convert '+' to space.
- (NSString*)gtm_stringByUnescapingFromURLArgument;

@end


#endif
