//
//  EVELinkedInClient.h
//  TestNSURLSession
//
//  Created by Tong Liu on 2/7/14.
//  Copyright (c) 2014 Evernote. All rights reserved.
//
//
//  Usage:
//  1. Check [EVELinkedInClient isAuthorized].
//  2. If not authorized, call [EVELinkedInClient authorizeWithPermissions:completion].
//  3. If auth is successful, perfom API calls using provided class method. Provide completion handlers if necessary.
//
//

#import <Foundation/Foundation.h>
#if TARGET_OS_IPHONE
#else
#import "ENLinkedInLoginWindowController.h"
#endif

@class EVELinkedInProfile;
@class EVEOAuthTokenItem;

typedef void (^EVELIAuthorizationCompletionHandler)(BOOL, NSError*);
typedef void (^EVELIProfileResultCompletionBlock)(EVELinkedInProfile*, NSDictionary*, NSError*);  // provide raw NSDictionary for backward-compatibility
typedef void (^EVELIArrayResultCompletionBlok)(NSArray*, NSDictionary*, NSError*);
typedef EVEOAuthTokenItem* (^EVELinkedInCredentialFetchBlock)(BOOL forceFetch);
typedef void (^EVELinkedInCredentialStoreBlock)(EVEOAuthTokenItem*);


@interface EVELinkedInClient : NSObject

#pragma mark - Authorization
/**
 *  Whether EVELinkedInClient is authorized with the service.
 *
 *  @return YES if either we have a cached, valid access token, or we successfully requested and stored a new token from Evernote service; NO otherwise.
 */
- (BOOL) isAuthorized;

/**
 *  Whether EVELinkedInClient has valid access token.
 *
 *  @return YES if we have a cached, valid access token; NO otherwise.
 */
- (BOOL) hasAccessToken;

/**
 *  Whether an authorization request is already in progress.
 *
 *  @return YES / NO
 */
- (BOOL) isAuthorizing;

/**
 *
 *  @return Cached access token item; nil if nonexistent.
 */
@property (strong) EVEOAuthTokenItem *tokenItem;

#if TARGET_OS_IPHONE
/**
 *  Start the authorization flow
 *
 *  @param permissions Permissions as defined in LinkedIn API "Member Permissions".
 *                     If not provided, will use "w_messages r_contactinfo r_network".
 *  @param completion  Invoked on the main thread after the operation is finished.
 */
- (void) authorizeWithPermissions:(NSArray *)permissions completion:(EVELIAuthorizationCompletionHandler)completion;
#else

/**
 *  Title for the dismiss button in linkedInLoginWindowController
 */
@property (nonatomic, copy) NSString *linkedInLoginDismissTitle;

/**
 *  @return The NSWindowController for LinkedIn authentication
 */
@property (nonatomic, strong, readonly) ENLinkedInLoginWindowController *linkedInLoginWindowController;

/**
 *  Start the authorization flow
 *
 *  @param permissions                    Permissions as defined in LinkedIn API "Member Permissions".
 *                                        If not provided, will use "w_messages r_contactinfo r_network".
 *  @param inParentWindowOrContainerView  If parameter is an NSWindow, the LinkedIn auth web view will run modally in the NSWindow,
 *                                        and the window will be dismissed by the dismissButton in ENLinkedInLoginWindowController.
 *                                        If parameter is an NSView, the LinkedIn auth web view will be added to the NSView,
 *                                        but the caller should be responsible for removing the view from the superView in the completion block.
 *                                        See the detailed reason in ENLinkedInLoginWindowController:73.
 *  @param completion                     Invoked on the main thread after the operation is finished.
 */
- (void) authorizeWithPermissions:(NSArray *)permissions inParentWindowOrContainerView:(id)parentWindowOrContainerView completion:(EVELIAuthorizationCompletionHandler)completion;
#endif

/**
 *  Make a call to https://api.linkedin.com/uas/oauth/invalidateToken. At the same time,
 *  remove locally cached access token, regardless of the API result.
 *  
 *  Important: the API call to https://api.linkedin.com/uas/oauth/invalidateToken doesn't work
 *  because official comments on API usage on certain website lie about 50% of the time ...
 *  As a result, what we do here is simply removing the access token from our local cache.
 *
 *  @param completion Invoked after the operation is finished.
 */
- (void) revokeAccessTokenWithCompletion:(EVELIAuthorizationCompletionHandler)completion;

/**
 *  Set Evernote username for storing the LinkedIn OAuth token.
 */
- (void) setEvernoteUsername:(NSString *)username;

// LinkedIn doesn't support direct API call to renew / refresh an OAuth 2 token yet!
// - (void) renewAccessTokenWithCompletion:(EVELIAuthorizationCompletionHandler)completion;

#pragma mark - Profile
/**
 *  Request profile for the authenticated user.
 *
 *  @param completion Invoked after the opration is finished.
 */
- (void) profileWithCompletion:(EVELIProfileResultCompletionBlock)completion;

/**
 *  Request a person's profile by LinkedIn member id.
 *
 *  @param memberId   The person's LinkedIn member id. Required.
 *  @param completion Invoked after the operation is finished.
 */
- (void) profileWithMemberId:(NSString *)memberId completion:(EVELIProfileResultCompletionBlock)completion;

/**
 *  Request a person's profile by public profile url.
 *
 *  @param url        The person's public profile url. Required.
 *  @param completion Invoked after the operation is finished.
 */
- (void) profileWithURL:(NSString *)url completion:(EVELIProfileResultCompletionBlock)completion;

/**
 *  Request a person's profile by the primary email
 *
 *  @param email      The person's primary email. Required.
 *  @param completion Invoked after the operation is finished.
 */
- (void) profileWithEmail:(NSString *)email completion:(EVELIProfileResultCompletionBlock)completion;

#pragma mark - Connections
/**
 *  Request connections basic info (only with memberId, formatted name and pictureUrl) for the current authenticated user.
 *
 *  @param completion Invoked after the operation is finished.
 *  @param date       If specified, this method will return the new connections since this date;
 *                    otherwise, all connectios will be returned.
 */
- (void) connectionsBasicInfoWithUpdateDateSince:(NSDate *)date completion:(EVELIArrayResultCompletionBlok)completion;

/**
 *  Request connections (a list of profiles) for the current authenticated user.
 *
 *  @param completion Invoked after the operation is finished.
 */
- (void) connectionsWithCompletion:(EVELIArrayResultCompletionBlok)completion;

/**
 *  Request a person's connections (a list of profiles) by LinkedIn member id.
 *
 *  @param memberId   The person's LinkedIn member id. Required.
 *  @param completion Invoked after the operation is finished.
 */
- (void) connectionsWithMemberId:(NSString *)memberId completion:(EVELIArrayResultCompletionBlok)completion;

#pragma mark - Messaging & Invitation
/**
 *  Send a private message to a person identified by LinkedIn member id.
 *
 *  @param memberId   The person's LinkedIn member id. Required.
 *  @param subject    The subject of the message. Required.
 *  @param body       The body of the message. Required.
 *  @param completion Invoked after the operation is finished.
 */
- (void) sendMessageToMemberId:(NSString *)memberId
                   withSubject:(NSString *)subject
                          body:(NSString *)body
                    completion:(void (^)(NSError *error))completion;

/**
 *  Send a "Invitation to connect" message to a person identified by the primary email.
 *
 *  @param email      The person's primary email. Required
 *  @param firstName  The person's first name. Required.
 *  @param lastName   The person's last name. Required.
 *  @param subject    The subject of the message. Required.
 *  @param body       The body of the message. Required.
 *  @param completion Invoked after the operation is finished.
 */
- (void) sendInvitationToEmail:(NSString *)email
                 withFirstName:(NSString *)firstName
                      lastName:(NSString *)lastName
                       subject:(NSString *)subject
                          body:(NSString *)body
                    completion:(void (^)(NSError *error))completion;

/**
 *  Send a "Invitation to connect" message to a person identified by LinkedIn member id.
 *
 *  @param memberId    The person's LinkedIn member id. Required.
 *  @param headerName  "Auth header name". Required. See http://developer.linkedin.com/documents/invitation-api
 *  @param headerValue "Auth header value". Required. See http://developer.linkedin.com/documents/invitation-api
 *  @param subject     The subject of the message. Required.
 *  @param body        The body of the message. Required.
 *  @param completion  Invoked after the operation is finished.
 */
- (void) sendInvitationToMemberId:(NSString *)memberId
      withAuthorizationHeaderName:(NSString *)headerName
                      headerValue:(NSString *)headerValue
                          subject:(NSString *)subject
                             body:(NSString *)body
                       completion:(void (^)(NSError *error))completion;

/**
 *  Update persisted and cached access token
 *
 *  @param tokenItem   The token item includes token, secret and expiration date
 */
- (void) updateAndStoreAccessToken:(EVEOAuthTokenItem *)tokenItem;

- (void) setLinkedInCredentialFetchBlock:(EVELinkedInCredentialFetchBlock)fetchBlock storeBlock:(EVELinkedInCredentialStoreBlock)storeBlock;

#pragma mark - Debugging
/**
 *  Make the cached access token invalid after given delay.
 *  This is to simulate token becoming invaliad / expired, after user revoke
 *  applicaiton's access from LinkedIn website.
 *  
 *  During testing, we find that after user revokes access on LinkedIn website,
 *  there's a ~20 min delay until token is actually invalidated. We also have
 *  troubles making our app re-appear in the application list on LinkedIn website.
 *
 *  tl;dr: It sucks.
 *
 *  @param seconds Specify a time interval, in seconds. If you specify 0 seconds,
 *  it'll get ignored and this call becomes a no-op.
 */
- (void) invalidateTokenAfterDelay:(NSTimeInterval)seconds;

/**
 *
 *  @return Number of seconds before token invalidation is performed.
 */
- (NSTimeInterval) secondsUntilTokenInvalidation;

@end


@interface EVELinkedInProfile : NSObject

@property (copy)     NSString *memberId;
@property (copy)     NSString *formattedName;
@property (copy)     NSString *firstName;
@property (copy)     NSString *lastName;
@property (copy)     NSString *headline;
@property (copy)     NSString *publicProfileUrl;
@property (copy)     NSString *pictureUrl;
@property (copy)     NSString *emailAddress;
@property (assign)   NSInteger distance;

@end


@interface EVEOAuthTokenItem : NSObject <NSCopying, NSCoding>

@property (copy) NSString *token;
@property (copy) NSString *tokenSecret;
@property (strong) NSDate *expirationDate;

- (id) initWithToken:(NSString *)token tokenSecret:(NSString *)tokenSecret expirationDate:(NSDate *)expirationDate;

@end
